/***************************************************************************
 *   Authors:                                                              *
 *   - Nicolas Brodu                                                       *
 *                                                                         *
 *   Copyrights:                                                           *
 *   - INRIA                                                               *
 *   - Nicolas Brodu (initial base code, some modifications in SVN)        *
 *                                                                         *
 *   This file is free software; you can redistribute it and/or modify     *
 *   it under the terms of the GNU General Public License as               *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Lesser General Public License for more details.                   *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/

#ifndef CLASSIFIER_AdaBoost
#define CLASSIFIER_AdaBoost

#include "ClassifierInterface.h"

/**

The AdaBoost class delegates training (including cross-validation) to another classifier.

It instanciates many copies of the same classifier and provides the same training set to each copy.
What change are the weights associated to the training instances: each classifier has its own set of weights.
The procedure is iterative, new weights for a new classifier copy are updated from the previous classification results.

Adaboost uses the base classifier class to collect all training instances
Then boosting occurs in the finishTraining() call.
Adaboost then calls prepare/train/finish cycles for the delegate classifier copies with different weights.

Be warned this function might be really time-consuming.
Adaboost takes as option the number of boosting steps to perform.

*/


struct AdaBoost : public ClassifierInterface {
    AdaBoost(ClassifierFactory* weakClassifierFactory, int numBoostingSteps, int resampleSize, int forceResampleFlag);
    ~AdaBoost();

    // Boosting is performed here. prepare/train/finish cycles are called on the delegate classifier copies with different weights.
    double finishTraining();
    // Uses the boosted combination of all delegated copies (N base classifier predict calls!)
    double predict(const itpp::vec& instance);

    // These are delegated to the target classifier
    bool setVFoldCrossValidation(int v);
    bool parseOption(int option, std::string option_arg);

    // Prepends Adaboost to the delegated name
    std::string getName();

    ClassifierFactory* factory;
    // resample is:
    // - 0 if using weights (no resampling)
    // - != 0 if using resampling
    //   - >0: resample size
    //   - -1: use training set size
    int nsteps, resample;
    std::vector<ClassifierInterface*> classifiers;
    // weights for the combination of all nsteps classifiers
    std::vector<double> combiWeights;
    double sumCombiWeights;
};

#endif
