/***************************************************************************
 *   Authors:                                                              *
 *   - Nicolas Brodu                                                       *
 *                                                                         *
 *   Copyrights:                                                           *
 *   - INRIA                                                               *
 *   - Nicolas Brodu (initial base code, some modifications in SVN)        *
 *                                                                         *
 *   This file is free software; you can redistribute it and/or modify     *
 *   it under the terms of the GNU General Public License as               *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Lesser General Public License for more details.                   *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/

#include <iostream>
#include <sstream>
#include <set>

#include <unistd.h>
#include <math.h>
#include "NearestNeighbours.h"

using namespace std;
using namespace itpp;

ClassifierInterface* NearestNeighboursFactory::instanciateClassifier() {
    switch (type) {
        case Euclidian: return new NearestNeighboursClassifier(&NearestNeighboursClassifier::EuclidianDistance, "Euclidian distance", false);
        case KolmogorovSmirnov: return new NearestNeighboursClassifier(&NearestNeighboursClassifier::KolmogorovSmirnovDistance, "Kolmogorov-Smirnov", true);
    };
    return 0;
}

NearestNeighboursFactory::NearestNeighboursFactory(NNType t) : type(t) {
}

NearestNeighboursFactory::~NearestNeighboursFactory() {
}


NearestNeighboursClassifier::NearestNeighboursClassifier(DistanceFunction dfunc, std::string distanceName, bool needSortSamples) : dist(dfunc), distName(distanceName), sortSamples(needSortSamples) {
    crossValidation = 0;
    Kp1 = 0;
    KStepCV = 0;
    KMaxCV = 0;
}

NearestNeighboursClassifier::~NearestNeighboursClassifier() {
}

//////////////////////////////////////////////////////////////////////////////
// Options
//////////////////////////////////////////////////////////////////////////////


ClassifierFactory::ClassifierInputType NearestNeighboursFactory::getType() {
    return (type==Euclidian) ? FeatureVector : DistributionSample;
}

std::string NearestNeighboursClassifier::getName() {
    return "Nearest Neighbours: " + distName;
}

int NearestNeighboursFactory::getMethodId() {
    return (type==Euclidian) ? 6 : 100;
}

std::string NearestNeighboursFactory::getClassifierOptions() {
    return "K:";
}

void NearestNeighboursFactory::printClassifierHelpText(ostream& os) {
    if (type==Euclidian) os<< "    6: Euclidian distance Nearest Neighbours (w)\n";
    if (type==KolmogorovSmirnov) os<< " 100: Kolmogorov-Smirnov Nearest Neighbours (w)\n";
os<<
"      -K <int>: (*) Number of neighbours to consider (default=3).\n"
;
}

bool NearestNeighboursClassifier::parseOption(int option, std::string option_arg) {
    if (option == 'K') {
        for (unsigned int i=0; i<option_arg.size(); ++i) if (option_arg[i]==',') option_arg[i]=' ';
        stringstream tokenizer(option_arg);
        tokenizer >> Kp1;
        tokenizer >> KMaxCV;
        tokenizer >> KStepCV;
        if ((Kp1<1) || (KStepCV<0) || (KStepCV && (KMaxCV+1<Kp1))) {
            cerr << "Invalid number of neighbours: " << optarg << endl;
            exit(1);
        }
        ++Kp1;
        return true;
    }
    return false;
}


//////////////////////////////////////////////////////////////////////////////
// classifier work
//////////////////////////////////////////////////////////////////////////////

double NearestNeighboursClassifier::EuclidianDistance(const itpp::vec& x, const itpp::vec&y) {
    vec delta = y - x;
    double sum = 0.0;
    for (int i=0; i<delta.size(); ++i) sum += delta(i) * delta(i);
    return sum;
    // elem_mult_sum is for ITPP >=v4.0
    //return elem_mult_sum(delta, delta);
}

double NearestNeighboursClassifier::KolmogorovSmirnovDistance(const itpp::vec& x, const itpp::vec&y) {
    double delta = 0.0;
    double maxdelta = 0.0;
    double cx = 0.0, cy = 0.0;
    for (int i=0; i<x.size(); ++i) {
        cx += x(i);
        cy += y(i);
        delta = fabs(cx - cy);
        if (delta > maxdelta) maxdelta = delta;
    }
    return maxdelta;
}

bool NearestNeighboursClassifier::setVFoldCrossValidation(int v) {
    crossValidation = v;
    return true;
}

int NearestNeighboursClassifier::getNumCVStep() {
    if (KStepCV) return KStepCV;
    return 11;
}

bool NearestNeighboursClassifier::supportWeightedInstance() {
    return true;
}

void NearestNeighboursClassifier::train(const itpp::vec& instance, double classification, double instanceWeight) {
    // parent maintains instance sets. Base prepareForTraining is OK, not overloaded here
    ClassifierInterface::train(instance, classification, instanceWeight);

    // Vector entries are samples from the same distribution
    // Sort them in order to get the density
    if (sortSamples) sort(trainSet.back());
}

double NearestNeighboursClassifier::finishTraining(int innercvstep) {

    int ntraininstances = trainSet.size();
    if (!ntraininstances) return 0.0;

    // cross-validate
    if (crossValidation && ((Kp1==0) || (KStepCV!=0))) {

        if (Kp1==0) {
            Kp1=2;
            KMaxCV=11;
            KStepCV = 6;
        }

        if (verbose) cout << "Cross-validating number of neighbours" << endl;
        vector<int> subsetBounds(crossValidation+1);
        for (int subset = 0; subset < crossValidation; ++subset) {
            subsetBounds[subset] = subset * ntraininstances / crossValidation;
        }
        subsetBounds[crossValidation] = ntraininstances; // end limit, after array
        set<int> cvK;
        if (KStepCV) {
            int rangeK = KMaxCV - Kp1+1;
            if (KStepCV>1) for (int ik = 0; ik < KStepCV; ++ik) {
                cvK.insert(Kp1 + (ik * rangeK) / (KStepCV - 1));
            }
            else cvK.insert((Kp1-1 + KMaxCV) / 2 +1);
        } else {
            for (int ik = 2; ik <= 11; ++ik) cvK.insert(ik);
        }
        int bestPerf = -1;
        int bestK = *cvK.begin();
        for (set<int>::const_iterator kit = cvK.begin(); kit!=cvK.end(); ++kit) {
            // unused loop if innercvstep<0
            for (int i=0; (i<innercvstep && kit!=cvK.end()); ++i) ++kit;

            int kcv = *kit;
            if (verbose) cout << "K=" << (kcv-1) << flush;
            int perf = 0;
            for (int subset = 0; subset < crossValidation; ++subset) {

                vector<int> instanceIdx;
                for (int i=0; i<subsetBounds[subset]; ++i) instanceIdx.push_back(i);
                for (int i=subsetBounds[subset+1]; i<ntraininstances; ++i) instanceIdx.push_back(i);

                for (int i=subsetBounds[subset]; i<subsetBounds[subset+1]; ++i) {
                    // Scores
                    if ( predictInternal(trainSet[i], kcv, instanceIdx) * trainClasses[i] > 0.0) ++perf;
                }
            }
            // favor largest number of neighbors for which perf is the same so as to help generalisation: use >= and not >
            if (perf>=bestPerf) {
                bestPerf = perf;
                bestK = kcv;
            }
            // now got cross-validation performance on the whole training set
            if (verbose) cout << ", perf=" << (perf*100.0/ntraininstances) << ", best perf=" << (bestPerf*100.0/ntraininstances) << endl;
            // cv step set externally ?
            if (innercvstep>=0) break;
        }
        if (verbose) cout << "Cross-validation selected using " << (bestK-1) << " nearest neighbours" <<endl;
        Kp1 = bestK;

        // Confidence in this classifier
        // If cross-validation, return the best perf as generalisation confidence
        return bestPerf;
    }

    // No cross-validation section here
    if (Kp1==0) Kp1 = 3+1; // default

    // confidence cannot be train set error as each point is its perfect neighbor
    // confidence unsupported => 1.0 (or equivalently this classifier thinks it classifies perfectly!)
    return 1.0;
}

double NearestNeighboursClassifier::predict(const vec& x) {
    int ntraininstances = trainSet.size();
    if (!ntraininstances) return 0.0;

    vector<int> tsetidx(ntraininstances);
    for (int i=0; i<ntraininstances; ++i) tsetidx[i] = i;

    return predictInternal(x, Kp1, tsetidx);
}

double NearestNeighboursClassifier::predictInternal(const itpp::vec& instance, int kp1orcv, std::vector<int>& tsetidx) {

    // look for the K+1 nearest neighbors
    // triplet: dist, class, instanceWeight
    vector< pair<double, pair<int, double> > > neighbors;
    for (vector<int>::iterator it = tsetidx.begin(); it!=tsetidx.end(); ++it) {
        int j = *it;
        if (trainWeights[j]<=0.0) continue; // ignore instances with null weights
        double d = (*dist)(instance,trainSet[j]);
        unsigned int k=neighbors.size();
        for (unsigned int l=0; l<neighbors.size(); ++l) if (d<neighbors[l].first) {k=l; break;}
        if (neighbors.size() < (unsigned int)kp1orcv) neighbors.resize(neighbors.size()+1);
        for (unsigned int l=neighbors.size()-1; l>k; --l) {
            neighbors[l] = neighbors[l-1];
        }
        if (k<neighbors.size()) neighbors[k] = make_pair(d, make_pair(((int)trainClasses[j] + 1)/2, trainWeights[j]) );
    }

    // OK, so we now have the k nearest neighbors
    // ... or not...
    if (neighbors.empty()) return 0.0; // then perf is not incremented
    // Normalize by the distance of the last neighbor (which exists at this point).
    double maxD = neighbors.back().first;
    if (maxD>0.0) {
        // pathological case: exact center of a circle of neighbors!
        if (neighbors[0].first == maxD) {
            // then give equal weight after kernel => consider them all at 0.0 dist, not 1.0 dist
            for (unsigned int k=0; k<neighbors.size(); ++k) neighbors[k].first = 0.0;
        }
        else for (unsigned int k=0; k<neighbors.size(); ++k) {
            neighbors[k].first /= maxD;
        }
    }

    // Now we have normalized distances between 0 and 1.
    // Apply a kernel so as to reverse the relation (0 is best dist, we want best coef = 1 after transform)
    // Use a simple triangular kernel here, linear w.r.t the the dist function.
    for (unsigned int k=0; k<neighbors.size(); ++k) {
        // Apply the kernel
        neighbors[k].first = 1.0 - neighbors[k].first;
        // Apply the instance weight
        neighbors[k].first *= neighbors[k].second.second;
    }
    // Can now vote by simple sum of coefs on the K, not K+1, nearest neighbors... if we have enough of them
    int maxK = (int)neighbors.size() < kp1orcv ? neighbors.size() : (kp1orcv-1);

    double votes[] = {0.0, 0.0};
    for (int k=0; k<maxK; ++k) {
        votes[neighbors[k].second.first] += neighbors[k].first;
    }
    // Tie? (including no vote case)
    if (votes[0]==votes[1]) {
        // Handles various pathological cases at the same time
        for (unsigned int k=0; k<neighbors.size(); ++k) if (neighbors[k].first>0.0) {
            votes[neighbors[k].second.first] += neighbors[k].first;
            break;
        }
        // really pathological if no draw breaker was found => no decision
        if (votes[0]==votes[1]) return 0.0;
    }

    return 1.0 - votes[0] * 2.0 / (votes[0] + votes[1]);
}



