/***************************************************************************
 *   Authors:                                                              *
 *   - Nicolas Brodu                                                       *
 *                                                                         *
 *   Copyrights:                                                           *
 *   - Nicolas Brodu                                                       *
 *   - INRIA                                                               *
 *                                                                         *
 *   This file is free software; you can redistribute it and/or modify     *
 *   it under the terms of the GNU General Public License as               *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Lesser General Public License for more details.                   *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/

#include "PredictionCombiners.h"

#include <vector>
#include <algorithm>
#include "math.h"

using namespace std;
using namespace itpp;

const int PredictionCombinerFactory::numRules = 7;


PredictionCombiner::PredictionCombiner(int ndim) {
}

PredictionCombiner::~PredictionCombiner() {
}

void PredictionCombiner::reset() {
}


struct SumPredictionCombiner : public PredictionCombiner {
    double totalWeights;
    vec predictions;

    SumPredictionCombiner(int ndim) : PredictionCombiner(ndim), totalWeights(0.0), predictions(ndim) {}

    void addPrediction(const vec& prediction, double confidence) {
        totalWeights += confidence;
        predictions += confidence * prediction;
    }

    vec combinePredictions() {
        if (totalWeights==0.0) return predictions;
        return predictions / totalWeights;
    }

    void reset() {totalWeights=0.0; predictions.zeros();}
};

struct ProductPredictionCombiner : public PredictionCombiner {
    vec predictions1;
    vec predictions2;

    ProductPredictionCombiner(int ndim) : PredictionCombiner(ndim), predictions1(ndim), predictions2(ndim) {
        predictions1 = 1.0;
        predictions2 = 1.0;
    }

    void addPrediction(const vec& prediction, double confidence) {
        for (int i=0; i<prediction.size(); ++i) {
            double p2 = (prediction(i) + 1.0) * 0.5; // 0 => class 1, 1 => class 2 : p is p2
            if (p2<0) p2=0;
            if (p2>1) p2=1;
            double p1 = 1-p2;
            predictions1(i) *= pow(p1,confidence);
            predictions2(i) *= pow(p2,confidence);
        }
    }

    vec combinePredictions() {
        vec ret(predictions1.size());
        for (int i=0; i<predictions1.size(); ++i) {
            double pt = predictions1(i) + predictions2(i);
            if (pt==0.0) ret(i) = 0.0;
            else ret(i) = (predictions2(i) * 2.0 / pt) - 1.0;
        }
        return ret;
    }

    void reset() {
        predictions1 = 1.0;
        predictions2 = 1.0;
    }

};

struct MedianPredictionCombiner : public PredictionCombiner {
    // for each dimension, all predictions
    std::vector<std::vector<double> > predictions1;
    std::vector<std::vector<double> > predictions2;

    MedianPredictionCombiner(int ndim) : PredictionCombiner(ndim), predictions1(ndim), predictions2(ndim) {}

    void addPrediction(const vec& prediction, double confidence) {
        for (int i=0; i<prediction.size(); ++i) {
            if (prediction(i)>0.0) predictions2[i].push_back(prediction(i) * confidence);
            else if (prediction(i)<0.0) predictions1[i].push_back(-prediction(i) * confidence);
        }
    }

    vec combinePredictions() {
        vec ret(predictions1.size());
        // sort, then get median
        for (unsigned int i=0; i<predictions1.size(); ++i) {
            std::sort(predictions1[i].begin(), predictions1[i].end());
            std::sort(predictions2[i].begin(), predictions2[i].end());
            if (predictions1[i].empty() && predictions2[i].empty()) {
                ret(i) = 0; continue;
            }
            if (predictions1[i].empty()) {
                ret(i) = predictions2[i][predictions2[i].size()/2]; continue;
            }
            if (predictions2[i].empty()) {
                ret(i) = -predictions1[i][predictions1[i].size()/2]; continue;
            }
            double median1 = predictions1[i][predictions1[i].size()/2];
            double median2 = predictions2[i][predictions2[i].size()/2];
            if (median1>median2) ret(i) = -median1;
            else if (median1<median2) ret(i) = median2;
            // else tie, => 0
        }
        return ret;
    }

    void reset() {
        for (unsigned int i=0; i<predictions1.size(); ++i) {
            predictions1[i].clear();
            predictions2[i].clear();
        }
    }
};

struct MajorityPredictionCombiner : public PredictionCombiner {
    // maintain weights to break the ties
    vec predictions1;
    vec predictions2;
    ivec counts1, counts2;

    MajorityPredictionCombiner(int ndim) : PredictionCombiner(ndim), predictions1(ndim), predictions2(ndim), counts1(ndim), counts2(ndim) {}

    void addPrediction(const vec& prediction, double confidence) {
        for (int i=0; i<prediction.size(); ++i) {
            if (prediction(i)>0.0) {
                ++counts2(i);
                predictions2(i) += prediction(i) * confidence;
            }
            else if (prediction(i)<0.0) {
                ++counts1(i);
                predictions1(i) -= prediction(i) * confidence;
            }
        }
    }

    vec combinePredictions() {
        vec ret(predictions1.size());
        for (int i=0; i<ret.size(); ++i) {
            if (counts1(i)>counts2(i)) ret(i) = -1.0;
            else if (counts1(i)<counts2(i)) ret(i) = 1.0;
            else {
                // break the tie.
                if (predictions1(i)>predictions2(i)) ret(i) = -1.0;
                else if (predictions1(i)<predictions2(i)) ret(i) = 1.0;
                // tie not broken, 0
            }
        }
        return ret;
    }

    void reset() {predictions1.zeros(); predictions2.zeros(); counts1.zeros(); counts2.zeros();}
};

struct WeightedMajorityPredictionCombiner : public PredictionCombiner {
    double totalWeights;
    vec predictions1;
    vec predictions2;

    WeightedMajorityPredictionCombiner(int ndim) : PredictionCombiner(ndim), totalWeights(0.0), predictions1(ndim), predictions2(ndim) {}

    void addPrediction(const vec& prediction, double confidence) {
        totalWeights += confidence;
        for (int i=0; i<prediction.size(); ++i) {
            // weighted vote
            if (prediction(i)>0.0) predictions2(i) += confidence;
            else if (prediction(i)<0.0) predictions1(i) += confidence;
        }
    }

    vec combinePredictions() {
        vec ret(predictions1.size());
        if (totalWeights==0.0) return ret; // null
        for (int i=0; i<ret.size(); ++i) {
            // weighted vote <0 or >0 anyway, but provide confidence as well
            if (predictions1(i)>predictions2(i)) ret(i) = -predictions1(i)/totalWeights;
            else if (predictions1(i)<predictions2(i)) ret(i) = predictions2(i)/totalWeights;
        }
        return ret;
    }

    void reset() {predictions1.zeros(); predictions2.zeros(); totalWeights=0.0;}
};

struct MaxDecisionPredictionCombiner : public PredictionCombiner {
    vec predictions;
    vec confs;

    MaxDecisionPredictionCombiner(int ndim) : PredictionCombiner(ndim), predictions(ndim), confs(ndim) {}

    void addPrediction(const vec& prediction, double confidence) {
        for (int i=0; i<prediction.size(); ++i) {
            if (fabs(prediction(i)) > fabs(predictions(i))) {
                predictions(i) = prediction(i);
                confs(i) = confidence;
            }
            else if (fabs(prediction(i)) == fabs(predictions(i))) {
                // use classifiers confidence only to break ties (fabs(-1) == fabs(1)...)
                if (confidence > confs(i)) {
                    predictions(i) = prediction(i);
                    confs(i) = confidence;
                }
                // if still ties, take first
            }
        }
    }

    vec combinePredictions() {
        return predictions;
    }

    void reset() {predictions.zeros(); confs.zeros();}
};

struct MaxConfidencePredictionCombiner : public PredictionCombiner {
    double maxConfidence; // start negative so the first observation is max even if it is alone
    double maxSumDecision;
    vec predictions;

    MaxConfidencePredictionCombiner(int ndim) : PredictionCombiner(ndim), maxConfidence(-1.0), maxSumDecision(-1.0), predictions(ndim) {}

    void addPrediction(const vec& prediction, double confidence) {
        if (confidence > maxConfidence) predictions = prediction;
        else if (confidence == maxConfidence) {
            // desisions only used to break ties
            double sumd = 0.0;
            for (int i=0; i<prediction.size(); ++i) sumd += fabs(prediction(i));
            if (sumd > maxSumDecision) {
                predictions = prediction;
                maxSumDecision = sumd;
            }
            // if still ties, take first
        }
    }

    vec combinePredictions() {
        return predictions;
    }

    void reset() {predictions.zeros(); maxConfidence=0.0; maxSumDecision=0.0;}
};


// Now create the factory

void PredictionCombinerFactory::printCombinerHelpText(ostream& os) {
os <<
"    1: Sum: Arithmetic weigthed average of each instance class decision by the classifier global confidence in the result.\n"
"    2: Product: Geometric weigthed average of each instance class decision by the classifier global confidence in the result.\n"
"    3: Median: for each instance, choose the class with maximal median value of the confidence*decision levels. This combination rule is less sensitive to large outliers than the above average rules.\n"
"    4: Majority: A vote occurs for each class. The class with maximum counts is returned. In case of a tie, the class with maximum overall confidence * decision level is returned.\n"
"    5: Weighted Majority: A vote occurs for each class, weighted by each classifier global confidence. The difference with Sum is that instance counts are used, not instance decision levels.\n"
"    6: Max decision: For each instance, consider only the class with max decision absolute value across all classifiers (ignores the classifiers confidence except to break ties).\n"
"    7: Max confidence: Consider only the classifier with maximal global confidence (ignores each specific instance decision level except to break ties).\n"
"\n"
"Terminology: an instance decision is the result of an individual classifier for that instance, the output of the \"predict\" function which is -1 or +1 or anything in between to indicate the accuracy of the prediction. A classifier confidence in the result is the global classifier confidence that it works well generally (ex: estimated generalisation accuracy, etc).\n"
"Example: -c3,1:2,1:8,2:100,3:100 applies Fisher discriminant analysis and Gaussian SVM on feature 1, and nearest neighbors using the Kolmogorov-Smirnov statistic with cumulated density estimation on features 2 and 3, then combines the results in a final prediction using the median rule.\n";
}

PredictionCombiner* PredictionCombinerFactory::instanciateCombiner(int combinerId, int ndim) {
    PredictionCombiner* ret = 0;
    switch(combinerId) {
        case 1: ret = new SumPredictionCombiner(ndim); break;
        case 2: ret = new ProductPredictionCombiner(ndim); break;
        case 3: ret = new MedianPredictionCombiner(ndim); break;
        case 4: ret = new MajorityPredictionCombiner(ndim); break;
        case 5: ret = new WeightedMajorityPredictionCombiner(ndim); break;
        case 6: ret = new MaxDecisionPredictionCombiner(ndim); break;
        case 7: ret = new MaxConfidencePredictionCombiner(ndim); break;
        default: return 0;
    }
    ret->rulenum = combinerId;
    return ret;
}

