/***************************************************************************
 *   Authors:                                                              *
 *   - Nicolas Brodu                                                       *
 *                                                                         *
 *   Copyrights:                                                           *
 *   - INRIA                                                               *
 *                                                                         *
 *   This file is free software; you can redistribute it and/or modify     *
 *   it under the terms of the GNU General Public License as               *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Lesser General Public License for more details.                   *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/

#ifndef PREDICTION_COMBINERS
#define PREDICTION_COMBINERS

#include <ostream>
#include <set>

#include <itpp/itbase.h>

// Handles the combination of different classifiers predictions
// Different combination rules have different needs (ex: median needs to store all values, max needs not)
struct PredictionCombiner {
    PredictionCombiner(int ndim);
    virtual ~PredictionCombiner();

    // prediction is a vector of size number of classes
    // It is double, not int, because methods produce continuous values that are then thresholded to get +1 / -1
    // confidence is the global confidence level of the prediction in 0..1
    virtual void addPrediction(const itpp::vec& prediction, double confidence) = 0;

    // Combine all observations. The returned vector is the same size as the predictions that were fed.
    // each member of the vector is a combination of the corresponding previous observations according to this combiner rule.
    // should not reset observations and be callable again after adding more observations (possibly 0)
    virtual itpp::vec combinePredictions() = 0;

    // prepare for a new set of combinations
    virtual void reset();

    // conditions that apply to this combiner
    // the caller may insert several conditions, identified by a number
    // and then check later on if a condition is satisfied before adding a prediction
    // Ex of condition: feature matches
    std::set<int> conditions;
    inline void addCondition(int cond) {conditions.insert(cond);}
    inline bool hasCondition(int cond) {return conditions.find(cond) != conditions.end();}
    
    // set by the factory
    int rulenum;
};

// Change to individual Factories, like methods?
// => no, keep it simple, no option to parse here, etc.
struct PredictionCombinerFactory {
    // Ids are written in the help text
    // returns 0 is ID not known
    static PredictionCombiner* instanciateCombiner(int combinerId, int ndim);
    // print an help text in the given output stream, with available IDs
    static void printCombinerHelpText(std::ostream& os);
    // ids are from 1 to numRules
    static const int numRules;
};


#endif
