/***************************************************************************
 *   Authors:                                                              *
 *   - Nicolas Brodu                                                       *
 *                                                                         *
 *   Copyrights:                                                           *
 *   - INRIA                                                               *
 *   - Nicolas Brodu (initial base code, some modifications in SVN)        *
 *                                                                         *
 *   This file is free software; you can redistribute it and/or modify     *
 *   it under the terms of the GNU General Public License as               *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Lesser General Public License for more details.                   *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/

#include <iostream>
#include <vector>
#include <set>
#include <unistd.h>

#include "SupportVectorMachines.h"

using namespace std;
using namespace itpp;

ClassifierInterface* SupportVectorMachinesFactory::instanciateClassifier() {
    return new SupportVectorMachinesClassifier(type);
}

SupportVectorMachinesClassifier::SupportVectorMachinesClassifier(libsvm::KernelType kerneltype) {
    svmCost = 0.0;
    svmGamma = 0.0; // set later on
    svmBias = 0.0;
    svmBiasIsSet = false;
    svmDegree = 0;
    svmCostMaxCV = 0.0;
    svmGammaMaxCV = 0.0;
    svmBiasMaxCV = 0.0;
    svmDegreeMaxCV = 0;
    svmCostStepCV = 0;
    svmGammaStepCV = 0;
    svmBiasStepCV = 0;
    svmDegreeStepCV = 0;

    // default values -- copied from svm-train.c
    svm_param.svm_type = libsvm::C_SVC;
    svm_param.gamma = -1.0;
    // default values
    svm_param.degree = 3;
    svm_param.C = 1.0;
    svm_param.coef0 = 0.0;
    svm_param.nu = 0.5;
    svm_param.cache_size = 100;
    svm_param.eps = 1e-3;
    svm_param.p = 0.1;
    svm_param.shrinking = 1;
    svm_param.probability = 1;
    svm_param.nr_weight = 0;
    svm_param.weight_label = 0;
    svm_param.weight = 0;
    svm_param.kernel_type = kerneltype;

    model = 0;
    problem = 0;
    featurenode = 0;

    crossValidation = 0;
}

SupportVectorMachinesClassifier::~SupportVectorMachinesClassifier() {

    if (featurenode) delete[] featurenode;
    if (problem) {
        delete problem->y;
        delete[] problem->x;
        delete problem;
    }
    if (model) svm_destroy_model(model);
}

SupportVectorMachinesFactory::SupportVectorMachinesFactory(libsvm::KernelType t) : type(t) {
}

SupportVectorMachinesFactory::~SupportVectorMachinesFactory() {
}


//////////////////////////////////////////////////////////////////////////////
// Options
//////////////////////////////////////////////////////////////////////////////


ClassifierFactory::ClassifierInputType SupportVectorMachinesFactory::getType() {
    return FeatureVector;
}

std::string SupportVectorMachinesClassifier::getName() {
    switch (svm_param.kernel_type) {
        case libsvm::LINEAR: return "Linear Support Vector Machine";
        case libsvm::RBF: return "Gaussian Support Vector Machine";
        case libsvm::POLY: return "Polynomial Support Vector Machine";
        case libsvm::SIGMOID: return "Sigmoid Support Vector Machine";
        default:;
    };
    return "Unknown Support Vector Machine";
}

int SupportVectorMachinesFactory::getMethodId() {
    switch (type) {
        case libsvm::LINEAR: return 7;
        case libsvm::RBF: return 8;
        case libsvm::POLY: return 9;
        case libsvm::SIGMOID: return 10;
        default:;
    };
    return 7;
}

std::string SupportVectorMachinesFactory::getClassifierOptions() {
    switch (type) {
        case libsvm::LINEAR: return "C:";
        case libsvm::RBF: return "C:G:";
        case libsvm::POLY: return "C:G:B:D:";
        case libsvm::SIGMOID: return "C:G:B:";
        default:;
    };
    return "";
}

void SupportVectorMachinesFactory::printClassifierHelpText(ostream& os) {
    switch (type) {
        case libsvm::POLY:
os <<
"    9: Polynomial Support Vector Machine\n"
"      -C <float>: (*) Cost factor of mistakes (default=1).\n"
"      -G <float>: (*) Gamma value, as in (gamma*u'*v + B)^degree. (default=1/nfeatures).\n"
"      -B <float>: (*) Constant value for the polynomial kernel. (default=0).\n"
"      -D <int>: (*) Degree of the polynomial kernel. (default=3).\n"
;
break;
        case libsvm::SIGMOID:
os <<
"   10: Sigmoid Support Vector Machine\n"
"      -C <float>: (*) Cost factor of mistakes (default=1).\n"
"      -G <float>: (*) Gamma value, as in tanh(gamma*u'*v + B). (default=1/nfeatures).\n"
"      -B <float>: (*) Bias value for the sigmoid kernel. (default=0).\n"
;
break;
        case libsvm::RBF:
os <<
"    8: Gaussian Support Vector Machine\n"
"      -C <float>: (*) Cost factor of mistakes (default=1).\n"
"      -G <float>: (*) Gamma value, as in exp(-gamma*|u-v|^2). (default=1/nfeatures).\n"
;
break;
        case libsvm::LINEAR:
os <<
"    7: Linear Support Vector Machine\n"
"      -C <float>: (*) Cost factor of mistakes (default=1).\n"
;
break;
        default:;
    };
}

bool SupportVectorMachinesClassifier::parseOption(int option, std::string option_arg) {
    if (option == 'C') {
        for (unsigned int i=0; i<option_arg.size(); ++i) if (option_arg[i]==',') option_arg[i]=' ';
        stringstream tokenizer(option_arg);
        tokenizer >> svmCost;
        tokenizer >> svmCostMaxCV;
        tokenizer >> svmCostStepCV;
        if ((svmCost<=0.0) || (svmCostStepCV<0) || (svmCostStepCV && (svmCostMaxCV<svmCost))) {
            cerr << "Invalid cost factor for the SVM: " << optarg << endl;
            exit(1);
        }
        return true;
    }

    if (option == 'G') {
        for (unsigned int i=0; i<option_arg.size(); ++i) if (option_arg[i]==',') option_arg[i]=' ';
        stringstream tokenizer(optarg);
        tokenizer >> svmGamma;
        tokenizer >> svmGammaMaxCV;
        tokenizer >> svmGammaStepCV;
        if ((svmGamma<=0.0) || (svmGammaStepCV<0) || (svmGammaStepCV && (svmGammaMaxCV<svmGamma))) {
            cerr << "Invalid gamma value for the SVM: " << optarg << endl;
            exit(1);
        }
        return true;
    }

    if (option == 'B') {
        for (unsigned int i=0; i<option_arg.size(); ++i) if (option_arg[i]==',') option_arg[i]=' ';
        stringstream tokenizer(optarg);
        tokenizer >> svmBias;
        tokenizer >> svmBiasMaxCV;
        tokenizer >> svmBiasStepCV;
        if ((svmBiasStepCV<0) || (svmBiasStepCV && (svmBiasMaxCV<svmBias))) {
            cerr << "Invalid bias value for the SVM: " << optarg << endl;
            exit(1);
        }
        svmBiasIsSet = true;
        return true;
    }

    if (option == 'D') {
        for (unsigned int i=0; i<option_arg.size(); ++i) if (option_arg[i]==',') option_arg[i]=' ';
        stringstream tokenizer(optarg);
        tokenizer >> svmDegree;
        tokenizer >> svmDegreeMaxCV;
        tokenizer >> svmDegreeStepCV;
        if ((svmDegree<1) || (svmDegreeStepCV<0) || (svmDegreeStepCV && (svmDegreeMaxCV<svmDegree))) {
            cerr << "Invalid polynomial degree value for the SVM: " << optarg << endl;
            exit(1);
        }
        return true;
    }

    return false;
}


//////////////////////////////////////////////////////////////////////////////
// classifier work
//////////////////////////////////////////////////////////////////////////////

bool SupportVectorMachinesClassifier::setVFoldCrossValidation(int v) {
    crossValidation = v;
    return true;
}

int SupportVectorMachinesClassifier::getNumCVStep() {
    if (!crossValidation) return 1;

    int nsteps = 1;

    if (svmCostStepCV) nsteps *= svmCostStepCV;
    else nsteps *= 5;
    if (svm_param.kernel_type==libsvm::LINEAR) return nsteps;

    if (svmGammaStepCV) nsteps *= svmGammaStepCV;
    else nsteps *= 5;
    if (svm_param.kernel_type==libsvm::RBF) return nsteps;

    if (svmBiasStepCV) nsteps *= svmBiasStepCV;
    else nsteps *= 5;
    if (svm_param.kernel_type==libsvm::SIGMOID) return nsteps;

    if (svmDegreeStepCV) nsteps *= svmDegreeStepCV;
    else nsteps *= 5;
    return nsteps;
}

bool SupportVectorMachinesClassifier::supportWeightedInstance() {
    return false;
}

double SupportVectorMachinesClassifier::finishTraining(int innercvstep) {

    int ntraininstances = trainSet.size();
    if (!ntraininstances) return 0.0;

    int nfeatures = trainSet[0].size();
    if (!nfeatures) return 0.0;

    // Reset parameter settings before cross-validation
    svm_param.degree = svmDegree;
    if (svmDegree==0) svm_param.degree = 3;
    svm_param.gamma = svmGamma;
    if (svmGamma==0.0) svm_param.gamma = 1.0 / nfeatures;
    svm_param.coef0 = svmBias; // may be null
    svm_param.C = svmCost;
    if (svmCost==0.0) svm_param.C = 1.0;


    if (featurenode) delete[] featurenode;
    if (problem) {
        delete problem->y;
        delete[] problem->x;
        delete problem;
    }
    if (model) svm_destroy_model(model);

    // Build the structures needed by libsvm
    featurenode = new libsvm::svm_node[ntraininstances*(nfeatures+1)];
    problem = new libsvm::svm_problem;
    problem->l = ntraininstances;
    problem->y = new double[ntraininstances];
    problem->x = new libsvm::svm_node*[ntraininstances];
    for (int i=0; i<ntraininstances; ++i) {
        problem->x[i] = &featurenode[i*(nfeatures+1)];
        problem->y[i] = trainClasses[i];
        for (int f=0; f<nfeatures; ++f) {
            problem->x[i][f].index = f+1;
        }
        problem->x[i][nfeatures].index = -1;
    }

    for (int i=0; i<ntraininstances; ++i) for (int f=0; f<nfeatures; ++f) problem->x[i][f].value = trainSet[i](f);
    const char* res = libsvm::svm_check_parameter(problem,&svm_param);
    if (res) {cerr << res << endl; exit(1);}

    int bestPerf = -1;

    // Enter cross-validation only if there is something to cross-validate!
    if (crossValidation && (
        ((svm_param.kernel_type==libsvm::LINEAR) && ((svmCost==0.0)||(svmCostStepCV!=0)))
    || ((svm_param.kernel_type==libsvm::RBF) && ((svmCost==0.0)||(svmCostStepCV!=0) || (svmGamma==0.0)||(svmGammaStepCV!=0)))
    || ((svm_param.kernel_type==libsvm::SIGMOID) && ((svmCost==0.0)||(svmCostStepCV!=0) || (svmGamma==0.0)||(svmGammaStepCV!=0) || (svmBiasIsSet==false)||(svmBiasStepCV!=0)))
    || ((svm_param.kernel_type==libsvm::POLY) && ((svmCost==0.0)||(svmCostStepCV!=0) || (svmGamma==0.0)||(svmGammaStepCV!=0) || (svmBiasIsSet==false)||(svmBiasStepCV!=0) || (svmDegree==0)||(svmDegreeStepCV!=0)))
    )) {

        // turn off probability estimations during cross-validation as recommended by libsvm
        svm_param.probability = 0;

        if (verbose) cout << "Cross-validating SVM" << endl;
        // numerical precision? we're doing grid search, so...
        double bestg = svmGamma;
        double bestc = svmCost;
        int bestDegree = svmDegree;
        double bestBias = svmBias;
        vector<double> target(ntraininstances);
        double cInterval = 1.0;
        double gInterval = 1.0;
        double bInterval = 1.0;

        set<double> cvCosts;
        // misclassification cost is a common parameter to all types
        if (svmCostStepCV) {
            if (svmCostStepCV>1) {
                double log10cMin = ::log10(svmCost);
                double log10cRange = ::log10(svmCostMaxCV) - log10cMin;
                cInterval = log10cRange / (svmCostStepCV-1.0);
                for (int ic=0; ic<svmCostStepCV; ++ic) {
                    cvCosts.insert(::pow10(log10cMin + ic * log10cRange / (svmCostStepCV - 1.0)));
                }
            }
            else {
                cvCosts.insert(::sqrt(svmCost*svmCostMaxCV)); // use geometric avg to be consistent
                cInterval = ::log10(svmCostMaxCV) - ::log10(svmCost);
            }
        }
        else {
            if (svmCost!=0.0) cvCosts.insert(svmCost);
            else {
                cvCosts.insert(0.01);
                cvCosts.insert(0.1);
                cvCosts.insert(1.0);
                cvCosts.insert(10.0);
                cvCosts.insert(100.0);
            }
        }

        set<double> cvGammas;
        if (svm_param.kernel_type!=libsvm::LINEAR) {
            if (svmGammaStepCV) {
                if (svmGammaStepCV>1) {
                    double log10gMin = ::log10(svmGamma);
                    double log10gRange = ::log10(svmGammaMaxCV) - log10gMin;
                    gInterval = log10gRange / (svmGammaStepCV-1.0);
                    for (int ig=0; ig<svmGammaStepCV; ++ig) {
                        cvGammas.insert(::pow10(log10gMin + ig * log10gRange / (svmGammaStepCV - 1.0)));
                    }
                }
                else {
                    cvGammas.insert(::sqrt(svmGamma*svmGammaMaxCV));
                    gInterval = ::log10(svmGammaMaxCV) - ::log10(svmGamma);
                }
            }
            else {
                if (svmGamma!=0.0) cvGammas.insert(svmGamma);
                else {
                    double invnf = 1.0/nfeatures;
                    cvGammas.insert(0.01*invnf);
                    cvGammas.insert(0.1*invnf);
                    cvGammas.insert(invnf);
                    cvGammas.insert(10.0*invnf);
                    cvGammas.insert(100.0*invnf);
                }
            }
        }
        else cvGammas.insert(-1.0); // flag

        set<int> cvDegree;
        if (svm_param.kernel_type==libsvm::POLY) {
            if (svmDegreeStepCV) {
                int rangeD = svmDegreeMaxCV - svmDegree;
                if (svmDegreeStepCV>1) for (int id = 0; id < svmDegreeStepCV; ++id) {
                    cvDegree.insert(svmDegree + (id * rangeD) / (svmDegreeStepCV - 1));
                }
                else cvDegree.insert((svmDegree + svmDegreeMaxCV) / 2);
            } else {
                if (svmDegree!=0) cvDegree.insert(svmDegree);
                else for (int id = 2; id <= 5; ++id) cvDegree.insert(id);
            }
        } else cvDegree.insert(-1); // flag

        set<double> cvBias; bool hasBias = false;
        if ((svm_param.kernel_type==libsvm::POLY) || (svm_param.kernel_type==libsvm::SIGMOID)) {
            if (svmBiasStepCV) {
                if (svmBiasStepCV>1) {
                    double rangeB = svmBiasMaxCV - svmBias;
                    bInterval = rangeB / (svmBiasStepCV - 1.0);
                    for (int ib=0; ib<svmBiasStepCV; ++ib) {
                        cvBias.insert(svmBias + ib * rangeB / (svmBiasStepCV - 1.0));
                    }
                }
                else {
                    cvBias.insert((svmBias + svmBiasMaxCV)*0.5);
                    bInterval = svmBiasMaxCV - svmBias;
                }
            }
            else {
                if (svmBias!=0.0) cvBias.insert(svmBias);
                else for (int ib=-2; ib<=2; ++ib) cvBias.insert(ib);
            }
            hasBias = true;
        } else {
            cvBias.insert(0.0);
        }

        // rough grid search first
        int cvstepcount=0;
        for (set<int>::const_iterator dit = cvDegree.begin(); dit!=cvDegree.end(); ++dit) {
            int degree = *dit;
            if (degree != -1) {
                svm_param.degree = degree;
                if (verbose) cout << "degree=" << degree << ", ";
            }
            for (set<double>::const_iterator bit = cvBias.begin(); bit!=cvBias.end(); ++bit) {
                double bias = *bit;
                if (hasBias) {
                    svm_param.coef0 = bias;
                    if (verbose) cout << "bias=" << svm_param.coef0 << ", ";
                }
                for (set<double>::const_iterator git = cvGammas.begin(); git!=cvGammas.end(); ++git) {
                    double gamma = *git;
                    if (gamma != -1.0) {
                        svm_param.gamma = gamma;
                        if (verbose) cout << "gamma=" << svm_param.gamma << ", ";
                    }
                    for (set<double>::const_iterator cit = cvCosts.begin(); cit!=cvCosts.end(); ++cit) {
                        if (innercvstep>=0 && cvstepcount++!=innercvstep) continue;
                        double cost = *cit;
                        if (cost != -1.0) {
                            svm_param.C = cost;
                            if (verbose) cout << "C=" << svm_param.C << flush;
                        }
                        svm_cross_validation(problem, &svm_param, crossValidation, &target[0]);
                        int perf = 0;
                        for (int i=0; i<ntraininstances; ++i) perf += (target[i] == trainClasses[i]);
                        // no need to normalize, all counts equal
                        if (perf > bestPerf) {
                            bestPerf = perf;
                            bestDegree = degree;
                            bestBias = bias;
                            bestg = gamma;
                            bestc = cost;
                        }
                        if (verbose) cout << ", perf= " << (perf*100.0/ntraininstances) << ", best perf= " << (bestPerf*100.0/ntraininstances) << endl;
                    }
                }
            }
        }

        // finer grid search now
        // misclassification cost is a common parameter to all types
        if ((svmCost==0.0)||(svmCostStepCV!=0)) {
            cvCosts.clear();
            double log10cMin = ::log10(bestc) - cInterval * 0.5;
            int nstepm1 = (svmCostStepCV>1) ? (svmCostStepCV - 1) : 5;
            for (int ic=0; ic<=nstepm1; ++ic) {
                cvCosts.insert(::pow10(log10cMin + ic * cInterval / nstepm1));
            }
        }

        if ((svm_param.kernel_type!=libsvm::LINEAR) && ((svmGamma==0.0)||(svmGammaStepCV!=0))) {
            cvGammas.clear();
            double log10gMin = ::log10(bestg) - gInterval * 0.5;
            int nstepm1 = (svmGammaStepCV>1) ? (svmGammaStepCV - 1) : 5;
            for (int ig=0; ig<=nstepm1; ++ig) {
                cvGammas.insert(::pow10(log10gMin + ig * gInterval / nstepm1));
            }
        }

        if (((svm_param.kernel_type==libsvm::POLY) || (svm_param.kernel_type==libsvm::SIGMOID)) && ((svmBiasIsSet==false)||(svmBiasStepCV!=0))) {
            cvBias.clear();
            double minb = bestBias - bInterval * 0.5;
            int nstepm1 = (svmBiasStepCV>1) ? (svmBiasStepCV - 1) : 5;
            for (int ib=0; ib<=nstepm1; ++ib) {
                cvBias.insert(minb + ib * bInterval / nstepm1);
            }
        }

        for (set<double>::const_iterator bit = cvBias.begin(); bit!=cvBias.end(); ++bit) {
            double bias = *bit;
            if (hasBias) {
                svm_param.coef0 = bias;
                if (verbose) cout << "bias=" << svm_param.coef0 << ", ";
            }
            for (set<double>::const_iterator git = cvGammas.begin(); git!=cvGammas.end(); ++git) {
                double gamma = *git;
                if (gamma != -1.0) {
                    svm_param.gamma = gamma;
                    if (verbose) cout << "gamma=" << svm_param.gamma << ", ";
                }
                for (set<double>::const_iterator cit = cvCosts.begin(); cit!=cvCosts.end(); ++cit) {
                    double cost = *cit;
                    if (cost != -1.0) {
                        svm_param.C = cost;
                        if (verbose) cout << "C=" << svm_param.C << flush;
                    }
                    svm_cross_validation(problem, &svm_param, crossValidation, &target[0]);
                    int perf = 0;
                    for (int i=0; i<ntraininstances; ++i) perf += (target[i] == trainClasses[i]);
                    // no need to normalize, all counts equal
                    if (perf > bestPerf) {
                        bestPerf = perf;
                        bestBias = bias;
                        bestg = gamma;
                        bestc = cost;
                    }
                    if (verbose) cout << ", perf= " << (perf*100.0/ntraininstances) << ", best perf= " << (bestPerf*100.0/ntraininstances) << endl;
                }
            }
        }

        svm_param.C = bestc;
        svm_param.gamma = bestg;
        svm_param.coef0 = bestBias;
        svm_param.degree = bestDegree;

        if (verbose) {
            cout << "Cross-validation selected";
            if ((svmCost==0.0)||(svmCostStepCV!=0)) cout << " C=" << svm_param.C;
            if ((svmGamma==0.0)||(svmGammaStepCV!=0)) cout << " gamma=" << svm_param.gamma;
            if ((svmDegree==0.0)||(svmDegreeStepCV!=0)) cout << " degree=" << svm_param.degree;
            if ((svmBiasIsSet==false)||(svmBiasStepCV!=0)) cout << " bias=" << svm_param.coef0;
            cout << endl;
        }
    }

    // also compute probability estimates during the final training phase
    // No, seems to be buggy (occasionally always return 0.5 0.5 proba)
    svm_param.probability = 1;

    // Done cross-validating best parameters!
    model = libsvm::svm_train(problem,&svm_param);

    // In case of cross-validation we might also return a confidence level
    // Otherwise, unsupported feature (this classifier is confident in itself => 1.0)
    return (bestPerf==-1) ? 1.0 : ((double)bestPerf / (double)ntraininstances);
}

double SupportVectorMachinesClassifier::predict(const vec& instance) {
    libsvm::svm_node x[instance.size()+1];
    for (int i = 0; i<instance.size(); ++i) {
        x[i].index = i+1;
        x[i].value = instance(i);
    }
    x[instance.size()].index = -1;
    double proba[2];
    double ret = libsvm::svm_predict_probability(model, x, proba);
//    cout << ret << " " << proba[0] << " " << proba[1] << endl;
    // probability estimates failed or cannot decide. Use non-probability version
    if (fabs(proba[0] - proba[1])<0.3) {
        ret = libsvm::svm_predict(model, x);
//        cout << ret << endl;
        return ret * 0.5; // say confidence in this result is 50%
    }
    // libsvm uses class 0 for +1 and class 1 for -1
    if (ret<0) return -proba[1];
    if (ret>0) return proba[0];
    return 0;
}


