/***************************************************************************
 *   Authors:                                                              *
 *   - Nicolas Brodu                                                       *
 *                                                                         *
 *   Copyrights:                                                           *
 *   - Nicolas Brodu (initial author of incremfa)                          *
 *   - INRIA (modifications/diffs in SVN)                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Lesser General Public License for more details.                   *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef UTILITY_H
#define UTILITY_H

/**
    Utility routines. Taken from Crogai project.

    @author Nicolas Brodu
*/
namespace Utility {

    void randomIsVerbose(bool verbosity);

    /** Random state holder object
        Taken from the streflop project (Nicolas Brodu)
        Declare one of this per thread, and use it as context to the random functions
        Object is properly set by the RandomInit functions
        This allows it to remain POD type
    */
    struct RandomState {
        // state vector
        unsigned int mt[624];
        int mti;
        // random seed that was used for initialization
        unsigned long seed;
    }
    #ifdef __GNUC__
    __attribute__ ((aligned (64))) // align state vector on cache line size
    #endif
    ;

    /// Default random state holder
    extern RandomState DefaultRandomState;



    /** Initialize the random number generator with the given seed.
        
        By default, the seed is taken from system time and printed out
        so the experiment is reproducible.

        You can set here a previous seed to reproduce it.
    
        This interface allows independance from the actual RNG used,
        and/or system functions.

        The RNG used is very precisely the Mersenne twister
        implementation by Agner Fog, released under GNU GPL at
        http://www.agner.org/random/randomc.htm.

        See also Utility.cpp and mersenne.cpp for information.
    */
    void RandomInit(long seed, RandomState& state = DefaultRandomState);        // re-seed

    long RandomInit(RandomState& state = DefaultRandomState);        // re-seed & return seed

    /// Returns the random seed that was used for the initialization
    long RandomSeed(RandomState& state = DefaultRandomState);

    /// Call this to set the pink noise generator number of rows
    /// 20 rows are used by default, or if this function is never called explicitly
    void PinkNoiseSetRows(int nrows = 20);

    /** Returns a random integer in the range [min, max], bounds
        included: min <= x <= max, from a uniform distribution.

        See comments in RandomInit.
    */
    int IRandom(int min, int max, RandomState& state = DefaultRandomState);       // output random integer

    /** Returns a random number between 0 (included) and 1 (excluded)
        0 <= x < 1 drawn from a uniform distribution.

        See comments in RandomInit. For the Mersene twister there are 32
        random bits in the returned number.
    */
    double Random(RandomState& state = DefaultRandomState);                     // output random float

    /** Return 32 random bits as a long */
    unsigned long BRandom(RandomState& state = DefaultRandomState);                    // output random bits

    /** Utility to get a number from a Normal distribution with
        mean 0, variance 1. Added to the other functions for this project
    */
    double NRandom(RandomState& state = DefaultRandomState);

    /** Returns a random number between 0 (included) and 1 (excluded) 0 <= x < 1
        The returned numbers have the "pink" noise property: power spectrum
        in 1/f, with f the frequency.
    */
    double PRandom(RandomState& state = DefaultRandomState);

    /// Fast pink noise approximation
    double PRandom2(RandomState& state = DefaultRandomState);
}

#endif

