/***************************************************************************
 *   Authors:                                                              *
 *   - Nicolas Brodu                                                       *
 *                                                                         *
 *   Copyrights:                                                           *
 *   - INRIA                                                               *
 *   - Nicolas Brodu (initial base code, some modifications in SVN)        *
 *   - Cephes Math Library Release 2.8:  June, 2000                        *
 *     Copyright 1984, 1987, 1992, 2000 by Stephen L. Moshier              *
 *                                                                         *
 *   This file is free software; you can redistribute it and/or modify     *
 *   it under the terms of the GNU General Public License as               *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Lesser General Public License for more details.                   *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/

#include "HyperMOG.h"

using namespace std;
using namespace itpp;

/*
Function from the Netlib Cephes mathematical library.
Adapted to compile stand-alone
*/
/*							psi.c
 *
 *	Psi (digamma) function
 *
 *
 * SYNOPSIS:
 *
 * double x, y, psi();
 *
 * y = psi( x );
 *
 *
 * DESCRIPTION:
 *
 *              d      -
 *   psi(x)  =  -- ln | (x)
 *              dx
 *
 * is the logarithmic derivative of the gamma function.
 * For integer x,
 *                   n-1
 *                    -
 * psi(n) = -EUL  +   >  1/k.
 *                    -
 *                   k=1
 *
 * This formula is used for 0 < n <= 10.  If x is negative, it
 * is transformed to a positive argument by the reflection
 * formula  psi(1-x) = psi(x) + pi cot(pi x).
 * For general positive x, the argument is made greater than 10
 * using the recurrence  psi(x+1) = psi(x) + 1/x.
 * Then the following asymptotic expansion is applied:
 *
 *                           inf.   B
 *                            -      2k
 * psi(x) = log(x) - 1/2x -   >   -------
 *                            -        2k
 *                           k=1   2k x
 *
 * where the B2k are Bernoulli numbers.
 *
 * ACCURACY:
 *    Relative error (except absolute when |psi| < 1):
 * arithmetic   domain     # trials      peak         rms
 *    DEC       0,30         2500       1.7e-16     2.0e-17
 *    IEEE      0,30        30000       1.3e-15     1.4e-16
 *    IEEE      -30,0       40000       1.5e-15     2.2e-16
 *
 * ERROR MESSAGES:
 *     message         condition      value returned
 * psi singularity    x integer <=0      MAXNUM
 */

/*
Cephes Math Library Release 2.8:  June, 2000
Copyright 1984, 1987, 1992, 2000 by Stephen L. Moshier
*/


inline double polevl( double x, double *coef, int N )
{
    double ans = *coef++;
    do
        ans = ans * x  +  *coef++;
    while( --N );
    return ans;
}

#include <limits>

#include <math.h>

static double A[] = {
 8.33333333333333333333E-2,
-2.10927960927960927961E-2,
 7.57575757575757575758E-3,
-4.16666666666666666667E-3,
 3.96825396825396825397E-3,
-8.33333333333333333333E-3,
 8.33333333333333333333E-2
};

#define EUL 0.57721566490153286061

//extern double PI, MAXNUM;
#define MAXNUM std::numeric_limits<double>::max()

#ifdef M_PI
#define PI M_PI
#else
#define PI 3.14159265358979323846
#endif


double psi(double x)
{
double p, q, nz, s, w, y, z;
int i, n, negative;

negative = 0;
nz = 0.0;

if( x <= 0.0 )
	{
	negative = 1;
	q = x;
	p = floor(q);
	if( p == q )
		{
		//mtherr( "psi", SING );
		return( MAXNUM );
		}
/* Remove the zeros of tan(PI x)
 * by subtracting the nearest integer from x
 */
	nz = q - p;
	if( nz != 0.5 )
		{
		if( nz > 0.5 )
			{
			p += 1.0;
			nz = q - p;
			}
		nz = PI/tan(PI*nz);
		}
	else
		{
		nz = 0.0;
		}
	x = 1.0 - x;
	}

/* check for positive integer up to 10 */
if( (x <= 10.0) && (x == floor(x)) )
	{
	y = 0.0;
	n = (int)x;
	for( i=1; i<n; i++ )
		{
		w = i;
		y += 1.0/w;
		}
	y -= EUL;
	goto done;
	}

s = x;
w = 0.0;
while( s < 10.0 )
	{
	w += 1.0/s;
	s += 1.0;
	}

if( s < 1.0e17 )
	{
	z = 1.0/(s * s);
	y = z * polevl( z, A, 6 );
	}
else
	y = 0.0;

y = log(s)  -  (0.5/s)  -  y  -  w;

done:

if( negative )
	{
	y -= nz;
	}

return(y);
}


void HyperMOG::init(int nclasses, int ndimensions) {
    v.set_size(nclasses);
    beta.set_size(nclasses);
    S.set_size(nclasses);
    lambda.set_size(nclasses);
    m.set_size(nclasses);
    classConst.set_size(nclasses);
    invS.set_size(nclasses);
    N = nclasses;
    D = ndimensions;
}

// update the class constants with the new parameters
void HyperMOG::prepareForEstimation() {
    double psi_sum_lambda = 0.0;
    for (int g=0; g<N; ++g) psi_sum_lambda += lambda(g);
    psi_sum_lambda = psi(psi_sum_lambda);
    for (int g=0; g<N; ++g) {
        classConst(g) = 0.5 *( D * log(2.0) - log(fabs(det(S(g)))) - D / beta(g) ) + psi(lambda(g)) - psi_sum_lambda;
        double psi_class_term = 0.0;
        for (int k=0; k<2; ++k) psi_class_term += psi(0.5*(v(g) + 1 - k));
        classConst(g) += 0.5 * psi_class_term;
        invS(g) = inv(S(g));
    }
}

double HyperMOG::estimatePosteriorLogLikelihood(int classNumber, const vec& x) {
    vec xdelta = x - m(classNumber);
    mat mat_term = (beta(classNumber) * xdelta.transpose()) * (invS(classNumber) * xdelta);
    return classConst(classNumber) - 0.5 * v(classNumber) * mat_term(0,0);
}

// Computes p(Class_i | x) -- should then be normalized
double HyperMOG::estimateUnNormalizedPosteriorProbability(int classNumber, const vec& x) {
    return trunc_exp(estimatePosteriorLogLikelihood(classNumber, x));
}

