function powscript(datasets,samplingFrequencies,methods,optargmethod)
% mkbp  Feature extraction of power in given frequency bands
%       Usage: powscript(datasets,Fs,methods)
%       Where:
%           datasets is an array of strings containing the name of the
%                    data sets to process. ex: {'s4','ovf'}
%           samplingFrequencies is an array with the sampling frequencies
%                   for each given data set. Ex: [125,512]
%           methods (optional) is an array of strings, containing the name of
%                   the methods to use for computing the spectral power
%                   Ex: {'iir','yuwa'}
%                   The default is to perform all methods.
%           optargmethod (optional) are extra arguments for each method
%       Available methods
%           pgrm: Periodogram computation using a raw fft on the signal
%           iir: Uses a butterworth filter to select each frequency band
%                the order is 4 by default and can be passed as optargmethod
%           yuwa: Yule-Walker fit of an autoregressive model, then used for
%                 estimating the power spectrum.
%                 The number of AR coefficients must be provided.
%           sgrm: Spectrogram (short-time Fourier transform)
%           wvd: Wigner-Ville distribution
%           mrlt: Morlet wavelet scalogram
%                 The wavelet support size must be provided as a coefficient of sqrt(nsamples)
%
% Uses the Time-Frequency Tool Box (tftb)
% Uses the octave-forge signal package
% Cross-validation is performed using least-square fitting

# sfiles=glob('data/s4trn/series*.txt');
# w=trn\trnclasses;
# score=sum((trn * w) .* trnclasses >0 )/size(trn,1)

if ((nargin<3) || any(ismember('all',methods(:)))),
    methods={'pgrm','iir','yuwa','sgrm','wvd','mrlt'};
end

if (nargin<4),
    optargmethod = zeros(1,6);
else
    resize(optargmethod,size(methods,2));
end

# number and position of frequency bands
# from [6 to 30 Hz), by increment of 1 Hz. bands are [f,f+1)
fvec = 4:1:30;
fcentervec = (fvec(1:end-1)+fvec(2:end))/2;
nfreq = length(fvec)-1;

mkdir pow;

for ds = 1:length(datasets),

    Fs = samplingFrequencies(ds);
    NyquistFreq = Fs/2;

    sfiles=glob(sprintf('data/%strn/series*.txt',datasets{ds}));
    ntrn=size(sfiles,1)/2;
    sfiles=glob(sprintf('data/%stst/series*.txt',datasets{ds}));
    ntst=size(sfiles,1)/2;

% NB20100720: datasets can now have a different series length, the simplest is to loop
for ser = 1:ntrn+ntst,

    trntst = 'trn';
    sernum = ser;
    if (ser>ntrn), trntst = 'tst'; sernum = ser - ntrn; end
    if sernum==1, appendflag=0;
    else appendflag=1;
    end

    data1=load(sprintf('data/%s%s/series%04d.txt',datasets{ds},trntst,sernum*2-1));
    data2=load(sprintf('data/%s%s/series%04d.txt',datasets{ds},trntst,sernum*2));
    nsamples=size(data1,1);
    assert(nsamples==size(data2,1));
    data = [data1 data2];
    classes = load(sprintf('data/%s%sclasses.txt',datasets{ds},trntst))';

    # Octave's support for nested/anonymous function is not quite there yet, missing closures.

    # Simple fft periodogram - no windowing
    if any(ismember('pgrm',methods(:))),
        nx = nsamples;
        X = fft(data);
        # DC f f f Nyq f f f : even case,  DC f f f f f f : odd case (Nyquist freq not there)
        X = X(1:(nx+rem(nx,2))/2, :);
        X = abs(X)/nx;              # non-duplicate modulus info, normalized
        X = X.^2;                   # power
        X(2:end-1+rem(nx, 2),:) = X(2:end-1+rem(nx, 2),:) * 2; # restore symmetric freqs in power
        f = (0:size(X)-1)*Fs/nx;    # frequency vector
        bands = zeros(nfreq,2);
        for k = 1:nfreq,
            bands(k,:) = log(sum( X((f>=fvec(k))&(f<fvec(k+1)),:), 1));
        end
        # gather both electrodes of each sample together
        feat_pgrm=reshape(bands,nfreq*2,1)';
        if (appendflag==1), dlmwrite(sprintf('pow/%s%s_pgrm.txt',datasets{ds},trntst),feat_pgrm,'-append','delimiter',' ');
        else dlmwrite(sprintf('pow/%s%s_pgrm.txt',datasets{ds},trntst),feat_pgrm,'delimiter',' ');
        end
    end

    # Butterworth IIR filter, energy summation in time
    # quite common in the BCI field, except here we split the frequency domain in small bands
    # Cross-validation is performed on the order parameter
    if any(ismember('iir',methods(:))),    
        [membership, idx] = ismember('iir',methods(:));
        horder=optargmethod(idx)/2;
        if (horder==0), horder=2; end
        %printf('iir butterworth order = %d\n',2*horder);
        assert(horder>1);        
        bands = zeros(nfreq,2);
        for k = 1:nfreq,
            [a,b]=butter(horder,[fvec(k) fvec(k+1)]/NyquistFreq);
            butpass = filter(a,b,data)(2*horder+1:nsamples,:);
            bands(k,:) = log(sum( butpass.^2, 1 ));
        end
        feat_iir = reshape(bands,nfreq*2,1)';
        if (appendflag==1), dlmwrite(sprintf('pow/%s%s_iir.txt',datasets{ds},trntst),feat_iir,'-append','delimiter',' ');
        else dlmwrite(sprintf('pow/%s%s_iir.txt',datasets{ds},trntst),feat_iir,'delimiter',' ');
        end
    end % of IIR method

    # Yule-Walker Auto-Regressive model parameter estimation, from which get a power density spectrum
    # needs cross-validation to get the AR size parameter (on the train set of course)
    if any(ismember('yuwa',methods(:))),
        [membership, idx] = ismember('yuwa',methods(:));
        nar=optargmethod(idx);
        %printf('Yule-Walker auto-regressive size = %d\n',nar);
        assert(nar>0);
        feat_yuwa = zeros(nfreq*2,1);
        for i = 1:2,
            feat_yuwa(nfreq*(i-1)+1:nfreq*i,1) = log(pyulear(data(:,i)',nar,fcentervec,Fs,'onesided'));
        end
        feat_yuwa = feat_yuwa';
        if (appendflag==1), dlmwrite(sprintf('pow/%s%s_yuwa.txt',datasets{ds},trntst),feat_yuwa,'-append','delimiter',' ');
        else dlmwrite(sprintf('pow/%s%s_yuwa.txt',datasets{ds},trntst),feat_yuwa,'delimiter',' ');
        end
    end % of yule-walker AR PSD estimation section

    # Spectrogram, the Time-Frequency toolbox is great and does all the work :)
    if any(ismember('sgrm',methods(:))),
        feat_sgrm = zeros(nfreq,2);
        for i = 1:2,
            [tfr,t,f] = tfrsp(data(:,i),1:nsamples,2^nextpow2(Fs));
            [margt,margf,E]=margtfr(tfr);
            # but we must still unfold negative frequencies
            # here we simply make all freqs positive so when summing energy later on we get both members of the freq pair.
            g=f; g(f<0)=-f(f<0);
            # and scale by the sampling frequency (the toolbox uses normalised freqs)
            g*=Fs;
            # now simply pick up the energy in each band
            for k = 1:nfreq,
                feat_sgrm(k,i) = log(sum( margf((g>=fvec(k))&(g<fvec(k+1))) ));
            end
        end
        feat_sgrm = reshape(feat_sgrm,nfreq*2,1)';
        if (appendflag==1), dlmwrite(sprintf('pow/%s%s_sgrm.txt',datasets{ds},trntst),feat_sgrm,'-append','delimiter',' ');
        else dlmwrite(sprintf('pow/%s%s_sgrm.txt',datasets{ds},trntst),feat_sgrm,'delimiter',' ');
        end
    end % of spectrogram section

    # Wigner-ville distribution using TFTB. See comments in the spectrogram section
    if any(ismember('wvd',methods(:))),
        feat_wvd = zeros(nfreq,2);
        for i = 1:2,
            [tfr,t,f] = tfrwv(data(:,i),1:nsamples,2^nextpow2(Fs));
            [margt,margf,E]=margtfr(tfr);
            # but we must still unfold negative frequencies
            # here we simply make all freqs positive so when summing energy later on we get both members of the freq pair.
            g=f; g(f<0)=-f(f<0);
            # and scale by the sampling frequency (the toolbox uses normalised freqs)
            g*=Fs;
            # now simply pick up the energy in each band
            for k = 1:nfreq,
                feat_wvd(k,i) = log(abs(sum( margf((g>=fvec(k))&(g<fvec(k+1))) )));
            end
        end
        feat_wvd = reshape(feat_wvd,nfreq*2,1)';
        if (appendflag==1), dlmwrite(sprintf('pow/%s%s_wvd.txt',datasets{ds},trntst),feat_wvd,'-append','delimiter',' ');
        else dlmwrite(sprintf('pow/%s%s_wvd.txt',datasets{ds},trntst),feat_wvd,'delimiter',' ');
        end
    end

    # Morlet wavelet scalogram using TFTB. See comments in the spectrogram section
    if any(ismember('mrlt',methods(:))),
        [membership, idx] = ismember('mrlt',methods(:));
        wsizefactor=optargmethod(idx);
        # wsize = tradeoff between freq and time resolution parameter
        # default in TFTB is sqrt(nsamples)
        if (wsizefactor==0), wsizefactor=1; end
        wsize = wsizefactor * sqrt(nsamples);
        %printf('Morlet wavelet size = %d\n',wsize);
        feat_morlet = zeros(nfreq,2);
        for i = 1:2,
            [tfr,t,f] = tfrscalo(data(:,i),1:nsamples,wsize,fvec(1)/Fs,fvec(length(fvec))/Fs,128);
            [margt,margf,E]=margtfr(tfr);
            # but we must still unfold negative frequencies
            # here we simply make all freqs positive so when summing energy later on we get both members of the freq pair.
            g=f; g(f<0)=-f(f<0);
            # and scale by the sampling frequency (the toolbox uses normalised freqs)
            g*=Fs;
            # now simply pick up the energy in each band
            for k = 1:nfreq,
                feat_morlet(k,i) = log(sum( margf((g>=fvec(k))&(g<fvec(k+1))) ));
            end
        end
        feat_morlet = reshape(feat_morlet,nfreq*2,1)';
        if (appendflag==1), dlmwrite(sprintf('pow/%s%s_mrlt.txt',datasets{ds},trntst),feat_morlet,'-append','delimiter',' ');
        else dlmwrite(sprintf('pow/%s%s_mrlt.txt',datasets{ds},trntst),feat_morlet,'delimiter',' ');
        end
    end

end # of series loop

end # of datasets loop
